# Suomilog
# Copyright (C) 2026 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
from collections import defaultdict
import pypykko.utils as pykko
from pypykko.reinflect import reinflect as pykko_reinflect
from pypykko.tokenizer import text2tokens as pykko_tokenize
from . import grammar

DICTIONARY: defaultdict[str, list[grammar.Token]] = defaultdict(list)

def tokenize(text: str) -> list[grammar.Token]:
	old_tokens: list[str] = pykko_tokenize(text)
	new_tokens: list[str] = []
	i = 0
	while i < len(old_tokens):
		if i < len(old_tokens) - 2 and old_tokens[i] == "\"" and old_tokens[i+2] == "\"":
			new_tokens.append("\"" + old_tokens[i+1] + "\"")
			i += 2
			continue

		if i < len(old_tokens) - 1 and old_tokens[i] == "\"":
			new_tokens.append("\"" + old_tokens[i+1])
			i += 1
			continue

		if i < len(old_tokens) - 1 and old_tokens[i+1] == "\"":
			new_tokens.append(old_tokens[i] + "\"")
			i += 1
			continue

		new_tokens.append(old_tokens[i])
		i += 1

	tokens: list[grammar.Token] = []
	token: str
	for token in new_tokens:
		tokenizer_bits = set()
		if token[:1] == "\"":
			token = token[1:]
			tokenizer_bits.add("-lquote")

		if token[-1:] == "\"":
			token = token[:-1]
			tokenizer_bits.add("-rquote")

		if token[:1] == "-":
			token = token[1:]
			tokenizer_bits.add("-lhyphen")

		if token[-1:] == "-":
			token = token[:-1]
			tokenizer_bits.add("-rhyphen")

		if token.strip() == "":
			continue

		alternatives = []
		for word in pykko.analyze(token):
			baseform, bits = baseformAndBits(word)
			bits |= tokenizer_bits
			alternatives.append((baseform, bits))

		# Jos sana löytyy suomilogin omasta sanakirjasta, lisää myös sieltä vaihtoehdot
		if token.lower() in DICTIONARY:
			alternatives += DICTIONARY[token.lower()]

		tokens.append(grammar.Token(token, alternatives))

	return tokens

def baseformAndBits(word: pykko.PykkoAnalysis) -> tuple[str, set[str]]:
	bits: set[str] = set()

	# Lisää kaikki morphtagit bitteinä
	morphtags: list[str] = re.split(r"(?=\+)", word.morphtags)
	bits |= set(morphtags)

	# TODO: Pitäisikö lisätä muitakin yhdistelmiä kuin tämä?
	if "+sg" in bits and "+nom" in bits:
		bits.add("+sg+nom")

	# Lisää sanaluokka, lemma ja pintamuoto bitteinä
	bits.add(f"{word.lemma}:{word.pos}")
	bits.add(f"{word.lemma}:")
	bits.add(f":{word.pos}")
	bits.add(f"«{word.wordform}»")

	return word.lemma, bits

SINGULAR_AND_PLURAL_CASES = [
	"+nom",
	"+gen",
	"+par",
	"+ess",
	"+abe",
	"+tra",
	"+ade",
	"+abl",
	"+all",
	"+ine",
	"+ela",
	"+ill",
]

PLURAL_CASES = [
	"+ins",
	"+com",
]

def inflect_nominal(word: str, plural_tag: str, case_tag: str, poss_tag: str = ""):
	assert plural_tag in ["+sg", "+pl"], plural_tag
	if plural_tag == "+pl":
		assert case_tag in SINGULAR_AND_PLURAL_CASES + PLURAL_CASES, f"{case_tag} not in {SINGULAR_AND_PLURAL_CASES + PLURAL_CASES}"
	
	else:
		assert case_tag in SINGULAR_AND_PLURAL_CASES, f"{case_tag} not in {SINGULAR_AND_PLURAL_CASES}"

	new_morphtags = []
	for analysis in pykko.analyze(word):
		morphtags: list[str] = re.split(r"(?=\+)", analysis.morphtags)
		new_morphtags = []
		is_nominal = False
		for tag in morphtags:
			if tag == "+conneg":
				is_nominal = False
				break

			if tag == "+sg" or tag == "+pl":
				new_morphtags.append(plural_tag)
				is_nominal = True
			
			elif tag in SINGULAR_AND_PLURAL_CASES or tag in PLURAL_CASES:
				new_morphtags.append(case_tag)
			
			else:
				new_morphtags.append(tag)
		
		if is_nominal:
			break
	
	else:
		new_morphtags = [plural_tag, case_tag]

	new_morphtags.append(poss_tag)

	return pykko_reinflect(word, "".join(new_morphtags))



# Suomilog
# Copyright (C) 2026 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .grammar import match_bits as match_bits
from .grammar import merge_bits as merge_bits

from .grammar import Token as Token

from .grammar import Grammar as Grammar

from .grammar import BaseRule as BaseRule
from .grammar import ProductionRule as ProductionRule

from .grammar import Nonterminal as Nonterminal

from .grammar import Output as Output
from .grammar import StringOutput as StringOutput

from .cykparser import CYKParser as CYKParser
from .cykparser import CYKAnalysis as CYKAnalysis




# Suomilog
# Copyright (C) 2026 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from typing import NamedTuple, Sequence
from . import grammar


type CYKTable = defaultdict[tuple[int, int], set[str]]
type SplitTable = defaultdict[tuple[int, int, str], set[int]]
type TokenOutputTable[OutputT] = defaultdict[tuple[int, int, str], set[OutputT]]


class CYKParser[OutputT]:
	token_rules: dict[str, grammar.Terminal | grammar.BaseRule[OutputT]]
	one_rules: defaultdict[str, set[str]]
	one_rules_expanded: defaultdict[str, set[str]]
	two_rules: defaultdict[tuple[str, str], set[str]]
	outputs: dict[tuple[str, str | tuple[str, str]], grammar.Output[OutputT] | "DenormalizeStartOutput[OutputT]" | "DenormalizeChainOutput[OutputT]" | "DenormalizeEndOutput[OutputT]"]

	def __init__(self, grammar: grammar.Grammar[OutputT], root_nonterminal_name: str):
		self.token_rules = {}
		self.one_rules = defaultdict(set)
		self.one_rules_expanded = defaultdict(set)
		self.two_rules = defaultdict(set)
		self.outputs = {}
		self.grammar = grammar
		self._to_CNF(root_nonterminal_name)

	def _to_CNF(self, root_nonterminal_name: str):
		_, expanded_grammar = self.grammar.expand_bits(root_nonterminal_name, set())
		for nonterminal_name in expanded_grammar:
			for rule in expanded_grammar[nonterminal_name]:
				# ProductionRule-luokka on niille säännöille, jotka voi jäsentää CYK-algoritmilla.
				# On myös sääntöjä, jotka perivät BaseRulen mutta eivät ProductionRulea. Tällöin luokka toteuttaa match-metodin, joka hoitaa jäsentämisen omalla tavallaan.
				if isinstance(rule, grammar.ProductionRule):
					new_words: list[str] = []
					is_nonterminal: list[bool] = []
					for word in rule.words:
						if isinstance(word, grammar.Terminal):
							self.token_rules[word.to_code()] = word
							new_words.append(word.to_code())
							is_nonterminal.append(False)
						
						elif isinstance(word, grammar.Nonterminal):
							new_words.append(word.name)
							is_nonterminal.append(True)
						
						else:
							assert False
					
					if len(new_words) == 1:
						self.one_rules[new_words[0]].add(nonterminal_name)
						assert (nonterminal_name, new_words[0]) not in self.outputs
						self.outputs[(nonterminal_name, new_words[0])] = rule.output
					
					else:
						prev = nonterminal_name
						j = id(rule)
						for i in range(len(new_words)-2):
							next = f"{nonterminal_name}_{j}_CONT{i}"
							pair = (new_words[i], next)
							self.two_rules[pair].add(prev)
							assert (prev, pair) not in self.outputs
							self.outputs[(prev, pair)] = DenormalizeChainOutput(is_nonterminal[i]) if i != 0 else DenormalizeEndOutput(is_nonterminal[i], rule.output)
							prev = next
						
						pair = (new_words[-2], new_words[-1])
						self.two_rules[pair].add(prev)
						assert (prev, pair) not in self.outputs, f"{prev}, {pair}"
						self.outputs[(prev, pair)] = DenormalizeStartOutput(is_nonterminal[-2], is_nonterminal[-1]) if len(new_words) > 2 else rule.output

				else:
					# Oma sääntö.
					# TODO: Tue muita kuin yhden saneen jäsentäviä omia sääntöjä.
					self.token_rules[nonterminal_name] = rule
		
		for a, b in list(self.one_rules.items()):
			self.one_rules_expanded[a] = set(b)
			queue = list(b)
			while queue:
				rule_name = queue.pop()
				for rule in self.one_rules[rule_name]:
					if rule not in self.one_rules_expanded[a]:
						self.one_rules_expanded[a].add(rule)
						queue.append(rule)
	
	def parse(self, tokens: list[grammar.Token]) -> "CYKAnalysis[OutputT]":
		cyk_table: CYKTable = defaultdict(set)
		split_table: SplitTable = defaultdict(set)
		token_outputs: TokenOutputTable = defaultdict(set)
		for i in range(len(tokens)):
			for rule_name, token_rule in self.token_rules.items():
				if isinstance(token_rule, grammar.BaseRule):
					if token_output := token_rule.match(self.grammar, tokens[i:i+1], set()):
						cyk_table[(i, i+1)] |= {rule_name} | self.one_rules_expanded[rule_name]
						token_outputs[(i, i+1, rule_name)] |= set(token_output)
					
				elif token_rule.matches_token(tokens[i]):
					cyk_table[(i, i+1)] |= {rule_name} | self.one_rules_expanded[rule_name]
		
		for span in range(2, len(tokens)+1):
			for start in range(len(tokens)-span+1):
				end = start + span
				for split in range(start+1, end):
					for rule1 in cyk_table[(start, split)]:
						for rule2 in cyk_table[(split, end)]:
							for rule_name in self.two_rules[(rule1, rule2)]:
								cyk_table[(start, end)] |= {rule_name} | self.one_rules_expanded[rule_name]
								split_table[(start, end, rule_name)] |= {split}
		
		return CYKAnalysis(self, tokens, cyk_table, split_table, token_outputs)
	
	def print(self):
		print("Token rules:")
		for a, b in self.token_rules.items():
			print(a, "<-", b.to_code())
		
		print("One rules:")
		for a, B in self.one_rules.items():
			for b in sorted(B):
				print(b, "<-", a)
		
		print("Two rules:")
		for (a, b), C in self.two_rules.items():
			for c in sorted(C):
				print(c, "<-", a, b)
		
		print("Expanded one rules:")
		for a, B in self.one_rules_expanded.items():
			print(B, "<-", a)
		
		print("Outputs:")
		for a, b in self.outputs.items():
			print(repr(a), repr(b))


class CYKAnalysis[OutputT]:
	def __init__(self, cyk_parser: CYKParser[OutputT], tokens: list[grammar.Token], cyk_table: CYKTable, split_table: SplitTable, token_outputs: TokenOutputTable):
		self.cyk_parser = cyk_parser
		self.tokens = tokens
		self.cyk_table = cyk_table
		self.split_table = split_table
		self.token_outputs = token_outputs

	def get_output(self, rule_name: str, start: int = 0, end: int = 0) -> frozenset[OutputT] | None:
		ans = self._get_output(rule_name, start, end)
		assert ans is None or all(not isinstance(arg, DenormalizedArgs) for arg in ans)
		return ans  # type: ignore

	def _get_output(self, rule_name: str, start: int = 0, end: int = 0) -> frozenset[OutputT | "DenormalizedArgs[OutputT]"] | None:
		if end <= 0:
			end = len(self.tokens) - end

		if rule_name not in self.cyk_table[(start, end)]:
			return frozenset()

		if not rule_name.startswith("."):  # jos kyseessä on terminaali
			return None

		ans: set[OutputT | DenormalizedArgs] = set()
		if token_output := self.token_outputs.get((start, end, rule_name), None):
			ans |= token_output

		for rule in self.cyk_table[(start, end)]:
			if output := self.cyk_parser.outputs.get((rule_name, rule), None):
				args = self.get_output(rule, start, end)
				assert args is not None
				assert isinstance(output, grammar.Output)
				for arg in args:
					ans.add(output.eval((arg,)))

		for split in self.split_table[(start, end, rule_name)]:
			for rule1 in self.cyk_table[(start, split)]:
				for rule2 in self.cyk_table[(split, end)]:
					if output := self.cyk_parser.outputs.get((rule_name, (rule1, rule2)), None):
						args1 = self._get_output(rule1, start, split)
						args2 = self._get_output(rule2, split, end)

						if args1 is None:
							args1 = frozenset({None})

						if args2 is None:
							args2 = frozenset({None})

						assert args1 and args2 and len(args1) > 0 and len(args2) > 0
						for arg1 in args1:
							assert not isinstance(arg1, DenormalizedArgs)
							for arg2 in args2:
								if isinstance(output, grammar.Output):
									assert not isinstance(arg2, DenormalizedArgs)
									ans.add(output.eval([arg for arg in (arg1, arg2) if arg is not None]))

								elif isinstance(output, DenormalizeStartOutput):
									assert not isinstance(arg2, DenormalizedArgs)
									ans.add(output.start_chain(arg1, arg2))

								elif isinstance(output, DenormalizeChainOutput):
									assert isinstance(arg2, DenormalizedArgs)
									ans.add(output.continue_chain(arg1, arg2))
								
								elif isinstance(output, DenormalizeEndOutput):
									assert isinstance(arg2, DenormalizedArgs)
									ans.add(output.end_chain(arg1, arg2))
		
		return frozenset(ans)

	def print(self) -> None:
		try:
			import rich

		except ModuleNotFoundError:
			print(self.cyk_table)
			return

		from rich.table import Table

		size = len(self.tokens)
		table = [[("" if row <= col else "X") for col in range(size)] for row in range(size)]
		for start in range(size):
			for end in range(start+1, size+1):
				table[end-start-1][end-1] = ", ".join(sorted(self.cyk_table[(start, end)]))
		
		rtable = Table(show_lines=True, show_footer=True)
		for token in self.tokens:
			rtable.add_column(token.surfaceform, repr(token))
		for row in reversed(table):
			rtable.add_row(*row)
		
		rich.print(rtable)


class DenormalizedArgs[OutputT](NamedTuple):
	args: tuple[OutputT, ...]


class DenormalizeStartOutput[OutputT]:
	def __init__(self, a_is_nonterminal: bool, b_is_nonterminal: bool):
		self.a_is_nonterminal = a_is_nonterminal
		self.b_is_nonterminal = b_is_nonterminal

	def __repr__(self):
		return "DenormalizeStartOutput()"

	def start_chain(self, a: OutputT | None, b: OutputT | None) -> DenormalizedArgs[OutputT]:
		if self.a_is_nonterminal and self.b_is_nonterminal:
			assert a is not None and b is not None
			return DenormalizedArgs((a, b))
		
		elif self.a_is_nonterminal:
			assert a is not None and b is None
			return DenormalizedArgs((a,))
		
		elif self.b_is_nonterminal:
			assert a is None and b is not None
			return DenormalizedArgs((b,))
		
		else:
			return DenormalizedArgs(())


class DenormalizeChainOutput[OutputT]:
	def __init__(self, a_is_nonterminal: bool):
		self.a_is_nonterminal = a_is_nonterminal

	def __repr__(self):
		return "DenormalizeChainOutput()"

	def continue_chain(self, arg: OutputT | None, args: DenormalizedArgs) -> DenormalizedArgs[OutputT]:
		if self.a_is_nonterminal:
			assert arg is not None
			return DenormalizedArgs((arg,) + args.args)
		
		else:
			assert arg is None
			return args


class DenormalizeEndOutput[OutputT]:
	def __init__(self, a_is_nonterminal: bool, output: grammar.Output[OutputT]):
		self.a_is_nonterminal = a_is_nonterminal
		self.output = output

	def __repr__(self):
		return "DenormalizeEndOutput(" + repr(self.output) + ")"

	def end_chain(self, arg: OutputT | None, args: DenormalizedArgs) -> OutputT:
		if self.a_is_nonterminal:
			assert arg is not None
			args = DenormalizedArgs((arg,) + args.args)
		
		else:
			assert arg is None
			args = args

		return self.output.eval(args.args)


# Suomilog
# Copyright (C) 2026 Iikka Hauhio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import AbstractSet, Callable, Mapping, Self, Sequence


def match_bits(tbits: AbstractSet[str], bits: AbstractSet[str]):
	positive_bits = {bit for bit in bits if not bit.startswith("!")}
	negative_bits = {bit[1:] for bit in bits if bit.startswith("!")}
	return tbits >= positive_bits and not (tbits & negative_bits)


def merge_bits(tbits: AbstractSet[str], bits: AbstractSet[str], extra_bitsets: Mapping[str, AbstractSet[str]] | None = None, allow_minus: bool = True) -> frozenset[str]:
	"""
	The first argument is the bits on the symbol (including $ and minus bits).
	The second argument is the incoming bits.

	Adds the incoming bits to tbits if tbits contains $.
	Removes any minus bits (bits preceded with a single backslash \\\\).

	For example:
	`merge_bits({"$", "\\\\+pl", "+sg"}, {"+pl", "+nom"}) == {"+sg", "+nom"}`
	"""
	new_bits: set[str] = set(tbits-{"$"}) | (bits if "$" in tbits else set())

	extra_bitsets = extra_bitsets or {}
	for bitset_name, bitset in extra_bitsets.items():
		if bitset_name in new_bits:
			new_bits |= bitset

	new_bits -= {bit for bit in new_bits if bit.startswith("$")}

	if allow_minus:
		minus_bits = {bit for bit in new_bits if bit.startswith("\\")}  # Removes the minus bits themselves
		minus_bits |= {minus_bit[1:] for minus_bit in minus_bits}  # Removes the matching plus bits

	else:
		minus_bits = set()

	return frozenset(new_bits - minus_bits)


class Token:
	"""
	Represents a token in the text to be parsed, typically a word or punctuation.

	A token is made of two components:
	1. The surfaceform, that is, the form in which the token is present in the tex
	2. A list of analysis alternatives consisting of a baseform and a set of bits.
	"""
	def __init__(self, surfaceform: str, alternatives: list[tuple[str, set[str]]]):
		self.surfaceform = surfaceform
		self.alternatives = alternatives

	def __repr__(self) -> str:
		return "Token(" + repr(self.surfaceform) + ", " + repr(self.alternatives) + ")"

	def __str__(self) -> str:
		return self.surfaceform + "[" + "/".join([baseform + "{" + ", ".join(bits) + "}" for baseform, bits in self.alternatives]) + "]"


class Terminal(ABC):
	"""
	The abstract base class of all terminal symbols.
	"""

	@abstractmethod	
	def to_code(self) -> str:
		...

	@abstractmethod
	def matches_token(self, token: Token) -> bool:
		...

	@abstractmethod
	def expand_bits(self, bits: AbstractSet[str]) -> Self:
		...


@dataclass(frozen=True)
class BaseformTerminal(Terminal):
	"""
	A terminal that matches to tokens that have at least one alternative with the given baseform and bits.
	"""
	baseform: str
	bits: AbstractSet[str]

	def to_code(self) -> str:
		return self.baseform + "{" + ",".join(self.bits) + "}"

	def matches_token(self, token: Token) -> bool:
		return any([tbf == self.baseform and match_bits(tbits, self.bits) for tbf, tbits in token.alternatives])

	def expand_bits(self, bits: AbstractSet[str]) -> "BaseformTerminal":
		new_bits = merge_bits(self.bits, bits)
		return BaseformTerminal(self.baseform, new_bits)


@dataclass(frozen=True)
class SurfaceformTerminal(Terminal):
	"""
	A terminal that matches to tokens with the correct surfaceform.
	"""
	surfaceform: str

	def to_code(self) -> str:
		return self.surfaceform

	def matches_token(self, token: Token) -> bool:
		return token.surfaceform == self.surfaceform

	def expand_bits(self, bits: AbstractSet[str]) -> "SurfaceformTerminal":
		return self


@dataclass(frozen=True)
class Nonterminal:
	"""
	A nonterminal that can appear on the right side of a production rule.
	"""
	name: str
	bits: AbstractSet[str]
	unexpanded: "Nonterminal | None" = None

	def to_code(self):
		return "." + self.name + ("{" + ",".join(self.bits) + "}" if self.bits else "")


type TerminalOrNonterminal = Terminal | Nonterminal


class Output[OutputT](ABC):
	"""
	A base class for rule outputs.
	"""

	@abstractmethod
	def eval(self, args: Sequence[OutputT]) -> OutputT:
		"""
		Constructs the output of the rule. `args` contains the outputs of the nonterminals in the rule.
		"""
		...


class StringOutput(Output[str]):
	"""
	A simple output method that treats the output code as a template string and subtitutes variables $1, $2, and so on with the outputs of the first nonterminal, the second nonterminal, and so on.
	"""
	def __init__(self, string: str):
		self.string = string

	def __repr__(self):
		return "StringOutput(" + repr(self.string) + ")"

	def eval(self, args):
		args = [arg.surfaceform if isinstance(arg, Token) else str(arg) for arg in args]
		ans = self.string
		for i, a in enumerate(args):
			var = "$"+str(i+1)
			if var not in ans:
				ans = ans.replace("$*", ",".join(args[i:]))
				break
			ans = ans.replace(var, str(a))
		return ans


class Grammar[OutputT]:
	rules: dict[str, list["BaseRule[OutputT]"]]
	"""
	A mapping from nonterminal names to rules.
	"""

	bitset_variables: dict[str, AbstractSet[str]]

	def __init__(self, rules: dict[str, list["BaseRule[OutputT]"]] | None = None, bitset_variables: dict[str, AbstractSet[str]] | None = None):
		self.rules = rules or {}
		self.bitset_variables = bitset_variables or {}

	def print(self):
		for nonterminal_name in sorted(self.rules):
			print(nonterminal_name, "::=")
			for rule in self.rules[nonterminal_name]:
				print(" " + rule.to_code())

	def copy(self):
		return Grammar(
			{name: rule.copy() for name, rule in self.rules.items()},
			self.bitset_variables.copy()
		)

	def update(self, grammar: "Grammar"):
		for nonterminal_name in grammar.rules:
			if nonterminal_name in self.rules:
				self.rules[nonterminal_name] += grammar.rules[nonterminal_name]
			else:
				self.rules[nonterminal_name] = grammar.rules[nonterminal_name].copy()

	def parse_grammar_line(self, line: str, output: Output[OutputT] | None = None, default_output: Callable[[str], Output[OutputT]] | None = StringOutput):
		if debug_level >= 1:
			print(line)

		tokens = line.replace("\t", " ").split(" ")
		if len(tokens) > 2 and tokens[0].startswith(".") and tokens[1] == "::=" and (output is not None or "->" in tokens):
			end = tokens.index("->") if "->" in tokens else len(tokens)
			nonterminal_name = tokens[0][1:]
			bits = set()
			if "{" in nonterminal_name and nonterminal_name[-1] == "}":
				i = nonterminal_name.index("{")
				bits = set(nonterminal_name[i+1:-1].split(","))
				nonterminal_name = nonterminal_name[:i]

			words: list[TerminalOrNonterminal] = []
			for token in tokens[2:end]:
				word = parse_word_in_grammar_line(token)
				if not word:
					continue
					
				words.append(word)

			if nonterminal_name not in self.rules:
				self.rules[nonterminal_name] = []

			if "->" in tokens:
				if output is not None or default_output is None:
					raise ValueError("Grammar line must have exactly one Output object. Ensure that `output` is not set if the grammar line has output code after an arrow `->`. If output code is present, the `default_output` must not be None.")

				output_code = " ".join(tokens[end+1:])
				output = default_output(output_code)

			assert output is not None
			rule = ProductionRule(nonterminal_name, words, output, bits)
			self.rules[nonterminal_name].append(rule)
			return rule

		else:
			raise ValueError("Syntax error on line `" + line + "'")

	def parse_variable_line(self, line: str):
		if debug_level >= 1:
			print(line)

		if debug_level >= 1:
			print(line)

		tokens = line.replace("\t", " ").split(" ")
		if len(tokens) == 3 and tokens[0].startswith("$") and tokens[1] == "=" and tokens[2].startswith("{") and tokens[2].endswith("}"):
			bitset_name = tokens[0]
			bits = set(tokens[2][1:-1].split(","))
			bits = merge_bits(bits, set(), self.bitset_variables, allow_minus=False)
			self.bitset_variables[bitset_name] = bits

		else:
			raise ValueError("Syntax error on line `" + line + "'")

	def expand_bits(self, nonterminal_name: str, bits: AbstractSet[str], extended: dict[str, list["BaseRule[OutputT]"]] | None = None) -> tuple[str, dict[str, list["BaseRule[OutputT]"]]]:
		if nonterminal_name not in self.rules:
			return nonterminal_name, {}

		ans: dict[str, list[BaseRule[OutputT]]] = extended or {}
		name = "." + nonterminal_name + "{" + ",".join(sorted(bits)) + "}"
		if name in ans:
			return name, ans

		ans[name] = []
		for rule in self.rules[nonterminal_name]:
			ans[name].append(rule.expand_bits(name, self, bits, ans))
		
		return name, ans


def parse_word_in_grammar_line(token: str) -> BaseformTerminal | SurfaceformTerminal | Nonterminal | None:
	if "{" in token and token[-1] == "}":
		has_bits = True
		bits = set(token[token.index("{")+1:-1].split(","))
		token = token[:token.index("{")]
	else:
		has_bits = False
		bits = set()

	if token == "":
		return None
	elif token.startswith(".") and token != ".":
		return Nonterminal(token[1:], bits)
	elif has_bits:
		return BaseformTerminal(token, bits)
	else:
		return SurfaceformTerminal(token)


debug_level = 0
indent = 0

def set_debug_level(n: int):
	global debug_level
	debug_level = n


class BaseRule[OutputT](ABC):
	"""
	Base class of all rules.

	Custom parsing algorithms can be implemented by overriding the `match` method that will be called for each range of tokens.
	"""
	@abstractmethod
	def to_code(self) -> str:
		...

	@abstractmethod
	def match(self, grammar: Grammar[OutputT], tokens: Sequence[Token], bits: AbstractSet[str]) -> list[OutputT]:
		"""
		The CYK parser calls this method for each range of tokens.
		If it returns a non-empty list, then this rule is considered to match the range.
		"""
		...

	@abstractmethod
	def expand_bits(self, name: str, grammar: Grammar[OutputT], bits: AbstractSet[str], extended: dict[str, list["BaseRule[OutputT]"]] | None = None) -> Self:
		...


class ProductionRule[OutputT](BaseRule[OutputT]):
	"""
	Represents a production rule. On the left side of the rule is a nonterminal name, on the right side is a list of terminals and nonterminals.
	"""

	nonterminal_name: str
	"""
	The left side of the rule: the name of the nonterminal that produces the right side.
	"""

	words: Sequence[TerminalOrNonterminal]
	"""
	The right side of the rule: a sequence of nonterminals and terminals this rule matches to.
	"""

	output: Output[OutputT]
	"""
	The Output object used to produce the output of this rule.
	"""

	bits: set[str]
	"""
	When expanded, this rule can only be expanded if $ matches these bits.
	"""

	def __init__(self, nonterminal_name: str, words: Sequence[TerminalOrNonterminal], output: Output[OutputT], bits: set[str] = set()):
		self.nonterminal_name = nonterminal_name
		self.words = words
		self.output = output
		self.bits = set(bits)
		self.positive_bits = set(bit for bit in bits if not bit.startswith("!"))
		self.negative_bits = set(bit[1:] for bit in bits if bit.startswith("!"))

	def __repr__(self):
		return "ProductionRule(" + repr(self.nonterminal_name) + ", " + repr(self.words) + ", " + repr(self.output) + ", bits=" + repr(self.bits) + ")"

	def to_code(self):
		ans = ""
		if self.bits:
			ans += "{" + ",".join(self.bits) + "} "
		ans += " ".join([w.to_code() for w in self.words])# + " -> " + repr(self.output)
		return ans

	def match(self, grammar: Grammar[OutputT], tokens: Sequence[Token], bits: AbstractSet[str]) -> list[OutputT]:
		raise NotImplementedError("Use the CYKParser to parse this rule")

	def expand_bits(self, name: str, grammar: Grammar[OutputT], bits: AbstractSet[str], extended: dict[str, list[BaseRule[OutputT]]] | None = None):
		positive_bits = set(bit for bit in bits if not bit.startswith("!"))
		negative_bits = set(bit[1:] for bit in bits if bit.startswith("!"))
		if not self.positive_bits <= positive_bits or self.negative_bits & positive_bits or self.positive_bits & negative_bits:
			return ProductionRule(name, [BaseformTerminal("<FALSE>", {"!"})], self.output)
		
		ans = []
		for word in self.words:
			if isinstance(word, Nonterminal):
				default_var = f"$default(.{word.name})"
				new_bits = merge_bits(word.bits|{default_var}, bits, grammar.bitset_variables)
				new_name, _ = grammar.expand_bits(word.name, new_bits, extended)
				ans += [Nonterminal(new_name, set(), unexpanded=word)]
			
			else:
				ans.append(word.expand_bits(bits))
		
		return ProductionRule(name, ans, self.output)